#ifndef _INC_MATRIX_
#define _INC_MATRIX_

#include "stdlib.h"

template<class T>
class SubMatrix;

template<class T>
class MatrixIterator;

enum accessType { ROW_MAJOR, COLUMN_MAJOR };

template<class T>
class Matrix {
protected:
	bool matrixFlag;
	accessType access;
	int numRows;
	int numCols;
	int length;
	int validILB;
	int validIUB;
	int validJLB;
	int validJUB;
	T* ptr;
public:
	Matrix(accessType access = ROW_MAJOR);
	Matrix(const Matrix& mat2);
	Matrix(int dimensionI, int dimensionJ, accessType access = ROW_MAJOR);
	~Matrix();
	accessType getAccess() const;
	int getRows() const;
	int getCols() const;
	int getLength() const;
	T* ptrElem(int indexI, int indexJ) const;
	T& operator()(int indexI, int indexJ) const;
	T operator()(float indexI, float indexJ) const;
	T getVal(int indexI, int indexJ) const;
	void setVal(int indexI, int indexJ, const T& val);
	Matrix<T>& set(T val);
	Matrix<T>& operator++();
	SubMatrix<T> sub(int posI, int posJ, int dimensionI, int dimensionJ) const;
	MatrixIterator<T> begin() const;
	bool isValid(int posI, int posJ) const;
};

template<class T>
class SubMatrix : public Matrix<T> {
public:
	SubMatrix();
	SubMatrix(const SubMatrix<T>& subMat2);
	SubMatrix(const Matrix<T>& mat, int posI, int posJ, int numRows, int numCols);
	~SubMatrix();
};

template<class T>
class MatrixIterator {
protected:
	bool active;
	int numRows;
	int numCols;
	int length;
	int indexRow;
	int indexCol;
	int validRowLB;
	int validRowUB;
	int validColLB;
	int validColUB;
	T* ptr;
public:
	MatrixIterator(const MatrixIterator<T>& matIt2);
	MatrixIterator(int numRows, int numCols, int length, int posI, int posJ, int validRowLB, int validRowUB, int validColLB, int validColUB, T* ptr);
	int getIndexRow() const;
	int getIndexCol() const;
	T& operator()() const;
	MatrixIterator<T> operator++();
	bool isActive() const;
	bool isValid() const;
};

template<class T>
inline Matrix<T>::Matrix(accessType access) : access(access) {
	matrixFlag = true;
	ptr = NULL;
	numRows = 0;
	numCols = 0;
	length = 0;
	validILB = 0;
	validIUB = 0;
	validJLB = 0;
	validJUB = 0;
}

template<class T>
Matrix<T>::Matrix(const Matrix& mat2) {
	matrixFlag = true;
	access = mat2.access;
	numRows = mat2.numRows;
	numCols = mat2.numCols;
	length = numCols;
	if(access == ROW_MAJOR) {
		validILB = 0;
		validIUB = numRows - 1;
		validJLB = 0;
		validJUB = numCols - 1;
	} else {
		validILB = 0;
		validIUB = numCols - 1;
		validJLB = 0;
		validJUB = numRows - 1;
	}
	ptr = new T[numRows * numCols];
	for(MatrixIterator<T> matIt = begin(), matIt2 = mat2.begin(); matIt2.isActive(); ++matIt, ++matIt2) {
		matIt() = matIt2();
	}
}

template<class T>
Matrix<T>::Matrix(int dimensionI, int dimensionJ, accessType access) : access(access) {
	matrixFlag = true;
	if(access == ROW_MAJOR) {
		numRows = dimensionI;
		numCols = dimensionJ;
		validILB = 0;
		validIUB = numRows - 1;
		validJLB = 0;
		validJUB = numCols - 1;
	} else {
		numRows = dimensionJ;
		numCols = dimensionI;
		validILB = 0;
		validIUB = numCols - 1;
		validJLB = 0;
		validJUB = numRows - 1;
	}
	length = numCols;
	ptr = new T[numRows * numCols];
}

template<class T>
Matrix<T>::~Matrix() {
	if(matrixFlag) if(ptr) delete []ptr;
}

template<class T>
inline accessType Matrix<T>::getAccess() const {
	return access;
}

template<class T>
inline int Matrix<T>::getRows() const {
	return numRows;
}

template<class T>
inline int Matrix<T>::getCols() const {
	return numCols;
}

template<class T>
inline int Matrix<T>::getLength() const {
	return length;
}

template<class T>
inline T* Matrix<T>::ptrElem(int indexI, int indexJ) const {
	if(access == ROW_MAJOR) return ptr + indexI * length + indexJ;
	else return ptr + indexJ * length + indexI;
}

template<class T>
inline T& Matrix<T>::operator()(int indexI, int indexJ) const {
	if(access == ROW_MAJOR) return ptr[indexI * length + indexJ];
	else return ptr[indexJ * length + indexI];
}

template<class T>
inline T Matrix<T>::operator()(float indexI, float indexJ) const {
	int intIndexI = (int) indexI;
	float fracIndexI = indexI - intIndexI;
	int intIndexJ = (int) indexJ;
	float fracIndexJ = indexJ - intIndexJ;
	T debug00 = getVal(intIndexI, intIndexJ);
	T debug10 = getVal(intIndexI + 1, intIndexJ);
	T debug01 = getVal(intIndexI, intIndexJ + 1);
	T debug11 = getVal(intIndexI + 1, intIndexJ + 1);
	T temp0 = fracIndexI * getVal(intIndexI, intIndexJ) + (1.0f - fracIndexI) * getVal(intIndexI + 1, intIndexJ);
	T temp1 = fracIndexI * getVal(intIndexI, intIndexJ + 1) + (1.0f - fracIndexI) * getVal(intIndexI + 1, intIndexJ + 1);
	T temp3 = fracIndexJ * temp0 + (1.0f - fracIndexJ) * temp1;
	return temp3;
}

template<class T>
inline T Matrix<T>::getVal(int indexI, int indexJ) const {
	if(access == ROW_MAJOR) return ptr[indexI * length + indexJ];
	else return ptr[indexJ * length + indexI];
}

template<class T>
inline void Matrix<T>::setVal(int indexI, int indexJ, const T& val) {
	if(access == ROW_MAJOR) ptr[indexI * length + indexJ] = val;
	else ptr[indexJ * length + indexI] = val;
}

template<class T>
Matrix<T>& Matrix<T>::set(T val) {
	for(MatrixIterator<T> matIt = begin(); matIt.isActive(); ++matIt) if(matIt.isValid()) matIt() = val;
	return (*this);
	return (*this);
}

template<class T>
Matrix<T>& Matrix<T>::operator++() {
	for(MatrixIterator<T> matIt = begin(); matIt.isActive(); ++matIt) if(matIt.isValid()) (matIt())++;
	return (*this);
}

template<class T>
inline SubMatrix<T> Matrix<T>::sub(int posI, int posJ, int dimensionI, int dimensionJ) const {
	SubMatrix<T> subMatrix(*this, posI, posJ, dimensionI, dimensionJ);
	return subMatrix;
}

template<class T>
inline MatrixIterator<T> Matrix<T>::begin() const {
	if(access == ROW_MAJOR) {
		MatrixIterator<T> matIt(numRows, numCols, length, 0, 0, validILB, validIUB, validJLB, validJUB, ptr);
		return matIt;
	} else {
		MatrixIterator<T> matIt(numRows, numCols, length, 0, 0, validJLB, validJUB, validILB, validIUB, ptr);
		return matIt;
	}
}

template<class T>
inline bool Matrix<T>::isValid(int posI, int posJ) const {
	if(posI < validILB) return false;
	if(posI > validIUB) return false;
	if(posJ < validJLB) return false;
	if(posJ > validJUB) return false;
	return true;
}

template<class T>
inline SubMatrix<T>::SubMatrix() {
	matrixFlag = false;
}

template<class T>
SubMatrix<T>::SubMatrix(const SubMatrix& subMat2) {
	matrixFlag = false;
	access = subMat2.access;
	numRows = subMat2.numRows;
	numCols = subMat2.numCols;
	length = subMat2.length;
	validILB = subMat2.validILB;
	validIUB = subMat2.validIUB;
	validJLB = subMat2.validJLB;
	validJUB = subMat2.validJUB;
	ptr = subMat2.ptr;
}

template<class T>
SubMatrix<T>::SubMatrix(const Matrix<T>& mat, int posI, int posJ, int dimension0, int dimension1) {
	matrixFlag = false;
	access = mat.getAccess();
	length = mat.getLength();
	if(access == ROW_MAJOR) {
		numRows = dimension0;
		numCols = dimension1;
		validILB = -posI;
		validIUB = mat.getRows() - posI - 1;
		validJLB = -posJ;
		validJUB = mat.getCols() - posJ - 1;
	} else {
		numRows = dimension1;
		numCols = dimension0;
		validILB = -posI;
		validIUB = mat.getCols() - posI - 1;
		validJLB = -posJ;
		validJUB = mat.getRows() - posJ - 1;
	}
	ptr = mat.ptrElem(posI, posJ);
}

template<class T>
inline SubMatrix<T>::~SubMatrix() { }

template<class T>
MatrixIterator<T>::MatrixIterator(const MatrixIterator<T>& matIt2) {
	active = matIt2.active;
	numRows = matIt2.numRows;
	numCols = matIt2.numCols;
	length = matIt2.length;
	indexRow = matIt2.indexRow;
	indexCol = matIt2.indexCol;
	validRowLB = matIt2.validRowLB;
	validRowUB = matIt2.validRowUB;
	validColLB = matIt2.validColLB;
	validColUB = matIt2.validColUB;
	ptr = matIt2.ptr;
}

template<class T>
MatrixIterator<T>::MatrixIterator(int numRows, int numCols, int length, int indexRow, int indexCol, int validRowLB, int validRowUB, int validColLB, int validColUB, T* ptr) {
	this->numRows = numRows;
	this->numCols = numCols;
	this->length = length;
	this->indexRow = indexRow;
	this->indexCol = indexCol;
	this->validRowLB = validRowLB;
	this->validRowUB = validRowUB;
	this->validColLB = validColLB;
	this->validColUB = validColUB;
	this->ptr = ptr;
	if(indexRow < 0) active = false;
	else if(indexCol < 0) active = false;
	else if (indexRow >= numRows) active = false;
	else if (indexCol >= numCols) active = false;
	else active = true;
}

template<class T>
inline int MatrixIterator<T>::getIndexRow() const {
	return indexRow;
}

template<class T>
inline int MatrixIterator<T>::getIndexCol() const {
	return indexCol;
}

template<class T>
inline T& MatrixIterator<T>::operator()() const {
	return *ptr;
}

template<class T>
inline MatrixIterator<T> MatrixIterator<T>::operator++() {
	indexCol++;
	if(indexCol >= numCols) {
		indexCol -= numCols;
		indexRow++;
		ptr += (length - numCols);
	}
	ptr++;
	if(indexRow >= numRows) active = false;
	return (*this);
}

template<class T>
inline bool MatrixIterator<T>::isActive() const {
	return active;
}

template<class T>
inline bool MatrixIterator<T>::isValid() const {
	if(indexRow < validRowLB) return false;
	if(indexRow > validRowUB) return false;
	if(indexCol < validColLB) return false;
	if(indexCol > validColUB) return false;
	return true;
}

#endif
